/*
 * MSC (MMC/SD controller) peripheral support.
 *
 * Copyright (C) 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include <l4/devices/clocks.h>
#include <l4/devices/cpm.h>
#include <l4/devices/dma.h>
#include <l4/devices/msc.h>
#include <l4/sys/types.h>
#include <stdint.h>



#ifdef __cplusplus

#include <l4/devices/hw_mmio_register_block.h>



// MMC/SD controller channel.

class Msc_channel
{
protected:
  l4_addr_t _msc_start;
  Hw::Register_block<32> _regs;
  l4_cap_idx_t _irq;
  Cpm_chip *_cpm;
  enum Clock_identifiers _clock;

  // Support eight cards.

  struct msc_card _cards[8];
  uint8_t _num_cards;
  int _card;

  // Utility methods.

  uint32_t get_field(uint32_t reg, uint32_t mask, uint8_t shift);
  void set_field(uint32_t reg, uint32_t mask, uint8_t shift, uint32_t value);

  // Low-level operations.

  void ack_irq(uint32_t flags);
  void unmask_irq(uint32_t flags);

  virtual void reset();

  bool set_clock_frequency(uint64_t frequency);
  void start_clock();
  void stop_clock();

  // Hardware properties.

  virtual bool have_dma_enable_in_command();
  virtual bool have_dma_selection();

  // Command properties.

  bool command_will_write(uint8_t index);
  bool command_with_data(uint8_t index);
  bool command_uses_busy(uint8_t index);
  uint8_t get_response_format(uint8_t index);

  bool app_command_will_write(uint8_t index);
  bool app_command_with_data(uint8_t index);
  bool app_command_uses_busy(uint8_t index);
  uint8_t get_app_response_format(uint8_t index);

  // Command initiation.

  bool send_app_command(uint8_t index, uint32_t arg);
  bool send_command(uint8_t index, uint32_t arg);
  bool send_command(uint8_t index, uint32_t arg, uint8_t response_format,
                    bool data, bool write, bool busy);

  // Response handling.

  bool have_response();
  void read_response(uint16_t *buffer, uint8_t units);
  bool wait_for_irq(uint32_t flags);
  bool wait_for_irq(uint32_t flags, unsigned int timeout);

  // Initialisation operations.

  bool check_sd();
  void init_sdio();
  void init_sdmem();
  void init_mmc();
  void identify_cards();
  void query_cards();

  // Transfer operations.

  uint32_t recv_data(struct dma_region *region, uint32_t count);
  uint32_t send_data(struct dma_region *region, uint32_t count);

  virtual uint32_t transfer(l4re_dma_space_dma_addr_t from_paddr,
                            l4re_dma_space_dma_addr_t to_paddr,
                            bool recv, uint32_t count) = 0;

public:
  explicit Msc_channel(l4_addr_t msc_start, l4_addr_t start, l4_cap_idx_t irq,
                       Cpm_chip *cpm, enum Clock_identifiers clock);

  virtual ~Msc_channel();

  void enable();

  msc_card *get_cards();

  uint8_t num_cards();

  uint32_t read(uint8_t card, struct dma_region *region,
                uint32_t address, uint32_t count);

  uint32_t read_blocks(uint8_t card, struct dma_region *region,
                       uint32_t block_address, uint32_t block_count);
};



// MMC/SD controller device control.

class Msc_chip
{
protected:
  l4_addr_t _msc_start, _start, _end;
  Cpm_chip *_cpm;

  virtual unsigned int num_channels() = 0;

public:
  explicit Msc_chip(l4_addr_t msc_start, l4_addr_t start, l4_addr_t end,
                    Cpm_chip *cpm);
};

#endif /* __cplusplus */
